import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Queue;

import javax.swing.JComboBox;
import javax.swing.JTextField;

import org.lekan.graphics.GraphicsProgram;
import org.lekan.graphics.SGImage;
import org.lekan.graphics.SGText;

/**
 * Who needs Google Maps when you have Pathfinder?
 * 
 * @author Martin Wickham
 *
 */
public class Pathfinder extends GraphicsProgram {
	
	public static void main(String[] args) {
		new Pathfinder();
	}
	
	// Constants
	//Files
	private static final String READ_FILE_NAME = "graph.txt";
	private static final String WRITE_FILE_NAME = "graph.txt";
	private static final String WP_PART = "Waypoint";
	private static final String DEST_PART = "Destination";
	private static final String EDGE_PART = "Edge";
	private static final String TAB = "\t";
	private static final String NEXT = "\n";
	
	//States
	private static final short DEST_STATE = 1;
	private static final short WP_STATE = 2;
	private static final short EDGE_STATE = 3;
	
	//Display
	private static final int CANVAS_WIDTH = 1000;
	private static final int CANVAS_HEIGHT = 618;
	private static final int DISPLAY_X = 5;
	private static final int DISPLAY_Y = 15;
	private static final int TEXT_WIDTH = 20;
//	private static final String DEST_TEXT_NAME= "Enter the name of a destination here";
//	private static final String EDGE_TEXT_NAME= "Enter the cost of the edge here";
//  TODO: implement better text box functionality
	
	//Buttons
	private static final String EDIT_TOGGLE_NAME = "Edit Mode";
	private static final String ADD_DEST_NAME = "Add Destination";
	private static final String ADD_WP_NAME = "Add Waypoint";
	private static final String ADD_EDGE_NAME = "Add Edge";
	private static final String SAVE_NAME = "Save";
	
	// [YOUR DATA STRUCTURES WILL GO HERE]
	private List<Waypoint> waypoints;
	private List<Destination> destinations;
	private List<Node> nodeList;
	private List<Edge> edges;
	
	// Variables
	
	//Data Finding
	private int nodeIndex;
	private int destIndex;
	private int wpIndex;
	
	//States
	private short state;
	private boolean canEdit;
	
	//Edge Dragging
	private boolean hasStartNode;
	private boolean hasEndNode;
	private Node startNode;
	private Node endNode;
	
	//Text
	private boolean hasEnteredName;
	private String name;
	private boolean hasEnteredCost;
	private String cost;
	
	//GUI
	private boolean hasMap;
	private SGImage map;
	private String imageName;
	
	/**
	 * Creates all buttons, the frame, and the title.
	 */
	@Override
	public void setup() {
		// Setup code goes here
		this.getFrame().setTitle("Pathfinder BETA!");
		this.getFrame().setDimensions(CANVAS_WIDTH, CANVAS_HEIGHT);
		String[] options = {ADD_DEST_NAME,ADD_WP_NAME,ADD_EDGE_NAME};
		addButton(EDIT_TOGGLE_NAME, NORTH);
		addComboBox(new JComboBox(options), NORTH);
		addTextField(new JTextField(TEXT_WIDTH), NORTH);
		addButton(SAVE_NAME, NORTH);
		hasStartNode = false;
		hasEndNode = false;
		hasEnteredName = false;
		hasEnteredCost = false;
		hasMap = false;
		nodeIndex = 0;
		destIndex = 0;
		wpIndex = 0;
		state = DEST_STATE;
		waypoints = new ArrayList<Waypoint>();
		destinations = new ArrayList<Destination>();
		nodeList = new ArrayList<Node>();
		edges = new ArrayList<Edge>();
	}
	
	@Override
	public void run() {
		if (!loadFile()) {
			System.err.println("An error occurred in loading the file.");
		}
		else {
			if (hasMap) {
				map = new SGImage(0, 0, imageName);
				this.getFrame().addObject(map);
			}
		}
		draw();
		// Run code goes here
		
	}
	
	/**
	 * Sets the state when a button is pressed.
	 */
	@Override
	
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals(EDIT_TOGGLE_NAME)) {
			canEdit = !canEdit;
			draw();
		}
		else if (e.getActionCommand().equals(SAVE_NAME)) {
			if (!saveFile()) {
				this.getFrame().drawText("An error occurred in saving the file.  See console for details.", DISPLAY_X, 50);
			}
			else {
				this.getFrame().drawText("File saved successfully!", DISPLAY_X, 50);
			}
		}
		else {
			JComboBox cb = (JComboBox)e.getSource();
	        String comboCommand = (String)cb.getSelectedItem();
	        if (!comboCommand.equals(ADD_WP_NAME) && 
	        		!comboCommand.equals(ADD_DEST_NAME) && // from text field
	        		!comboCommand.equals(ADD_EDGE_NAME) &&
	        		state == EDGE_STATE) {
	        	hasEnteredCost = true;
	        	cost = e.getActionCommand();
	        	System.out.println(cost);
	        }
	        else if (!comboCommand.equals(ADD_WP_NAME) && 
	        		!comboCommand.equals(ADD_DEST_NAME) && // from text field
	        		!comboCommand.equals(ADD_EDGE_NAME) &&
	        		state == DEST_STATE) {
	        	hasEnteredName = true;
	        	name = e.getActionCommand();
	        	System.out.println(name);
	        }
	        else if (canEdit){
	        	if (comboCommand.equals(ADD_DEST_NAME)) {
	        		draw();
	        		state = DEST_STATE;
	        	}
	        	else if (comboCommand.equals(ADD_WP_NAME)) {
	        		draw();
	        		state = WP_STATE;
	        	}
	        	else if (comboCommand.equals(ADD_EDGE_NAME)) {
	        		draw();
	        		state = EDGE_STATE;
	        	}
	        }
	        else {
	        	//if (e.getActionCommand().equals(FIND_PATH_NAME)) {
	        	//	findPath(pathStartNode);
	        	//	drawPath();
	        	//TODO: implement pathfinding GUI
	        	//}
	        	/*else*/ if (comboCommand.equals(ADD_DEST_NAME)) {
	        		draw();
	        		this.getFrame().addObject(new SGText("Must be in edit mode to add a Destination.", DISPLAY_X, DISPLAY_Y));
	        		state = DEST_STATE;
	        	}
	        	else if (comboCommand.equals(ADD_WP_NAME)) {
	        		draw();
	        		this.getFrame().addObject(new SGText("Must be in edit mode to add a Waypoint.", DISPLAY_X, DISPLAY_Y));
	        		state = WP_STATE;
	        	}
	        	else if (comboCommand.equals(ADD_EDGE_NAME)) {
	        		draw();
	        		this.getFrame().addObject(new SGText("Must be in edit mode to add an Edge.", DISPLAY_X, DISPLAY_Y));
	        		state = EDGE_STATE;
	        	}
	        }
		}
	}
	
	/**
	 * Does nothing (Yet)
	 */
	@Override
	public void mouseClicked(MouseEvent e) {
//		// :D
//		// (Replace this with your code)
//		this.getFrame().drawText(":D", new Random().nextInt(CANVAS_WIDTH), new Random().nextInt(CANVAS_HEIGHT));
	}
	
	/**
	 * If drawing an edge, this method will set the start node if isNearNode().
	 */
	@Override
	public void mousePressed(MouseEvent e) {
		int xPos = e.getX();
		int yPos = e.getY();
		switch (state) {
		case EDGE_STATE:
			isNearNode(xPos, yPos);
			break;
		}
	}
	
	/**
	 * If drawing an edge, this method will render the edge as a line.
	 */
	@Override
	public void mouseDragged(MouseEvent e) {
		draw();
		int xPos = e.getX();
		int yPos = e.getY();
		switch(state) {
		case EDGE_STATE:
			if (hasStartNode) {
				this.getFrame().drawLine(startNode.getX(), startNode.getY(), xPos, yPos); 
			}
		}
		
	}
	
	/**
	 * Depending on the state, this method will do one of the following:<br />
	 * (0) do nothing<br />
	 * (1) make a new destination at the coordinates of the MouseEvent<br /> 
	 * (2) make a new waypoint at the coordinates of the MouseEvent<br />
	 * (3) make a new edge between startNode and endNode (if applicable)<br />
	 */
	@Override
	public void mouseReleased(MouseEvent e) {
		draw();
		int xPos = e.getX();
		int yPos = e.getY();
		switch(state) {
		case DEST_STATE:
			if (hasEnteredName) {
				makeDestination(xPos, yPos, name);
				hasEnteredName = false;
			}
			else {
				makeDestination(xPos, yPos);
			}
			break;
		case WP_STATE:
			makeWaypoint(xPos, yPos);
			break;
		case EDGE_STATE:
			if (hasStartNode && isNearNode(xPos, yPos)) {
				if (hasEnteredCost) {
					makeEdge(startNode, endNode, cost);
					hasEnteredCost = false;
				}
				else {
					makeEdge(startNode, endNode);
				}
				
			}
			hasStartNode = false;
			hasEndNode = false;
			break;
		}
	}
	
	private void makeEdge(Node startNode, Node endNode, String cost) {
		if (startNode == endNode) {
			System.out.println("You can't connect a node to itself!");
		}
		else {
			edges.add(new Edge(startNode, endNode, Double.parseDouble(cost)));
			draw();
		}
	}

	/**
	 * Makes a new edge between Node start and node End
	 * @param start the first Node to connect to
	 * @param end the end Node to connect to
	 */
	public void makeEdge(Node start, Node end) {
		if (start == end) {
			System.out.println("You can't connect a node to itself!");
		}
		else {
			edges.add(new Edge(start, end));
			draw();
		}
	}
	
	/**
	 * Makes a new waypoint at (xPos, yPos).
	 * @param xPos the x coordinate of the new waypoint
	 * @param yPos the y coordinate of the new waypoint
	 */
	public void makeWaypoint(int xPos, int yPos) {
		nodeList.add(new Waypoint(xPos, yPos, wpIndex, nodeIndex));
		waypoints.add(new Waypoint(xPos, yPos, wpIndex, nodeIndex));
		wpIndex++;
		nodeIndex++;
		draw();
	}
	
	/**
	 * Makes a new destination at (xPos, yPos) using the given name.
	 * @param xPos the x coordinate of the new destination
	 * @param yPos the y coordinate of the new destination
	 * @param name the name of the new destination
	 */
	public void makeDestination(int xPos, int yPos, String name) {
		nodeList.add(new Destination(xPos, yPos, name, destIndex, nodeIndex));
		destinations.add(new Destination(xPos, yPos, name, destIndex, nodeIndex));
		destIndex++;
		nodeIndex++;
		draw();
	}
	
	/**
	 * Makes a new destination at (xPos, yPos) using the default name.
	 * @param xPos the x coordinate of the new destination
	 * @param yPos the y coordinate of the new destination
	 */
	public void makeDestination(int xPos, int yPos) {
		nodeList.add(new Destination(xPos, yPos, destIndex, nodeIndex));
		destinations.add(new Destination(xPos, yPos, destIndex, nodeIndex));
		destIndex++;
		nodeIndex++;
		draw();
	}
	
	private void loadEdge(Node startNode, Node endNode, String cost) {
		if (startNode == endNode) {
			System.out.println("You can't connect a node to itself!");
		}
		else {
			edges.add(new Edge(startNode, endNode, Double.parseDouble(cost)));
		}
		
	}

	/**
	 * Makes a new edge between Node start and node End
	 * @param start the first Node to connect to
	 * @param end the end Node to connect to
	 */
	public void loadEdge(Node start, Node end) {
		if (start == end) {
			System.out.println("You can't connect a node to itself!");
		}
		else {
			edges.add(new Edge(start, end));
		}
	}
	
	/**
	 * Makes a new waypoint at (xPos, yPos).
	 * @param xPos the x coordinate of the new waypoint
	 * @param yPos the y coordinate of the new waypoint
	 */
	public void loadWaypoint(int xPos, int yPos) {
		nodeList.add(new Waypoint(xPos, yPos, wpIndex, nodeIndex));
		waypoints.add(new Waypoint(xPos, yPos, wpIndex, nodeIndex));
		wpIndex++;
		nodeIndex++;
	}
	
	/**
	 * Makes a new destination at (xPos, yPos) using the given name.
	 * @param xPos the x coordinate of the new destination
	 * @param yPos the y coordinate of the new destination
	 * @param name the name of the new destination
	 */
	public void loadDestination(int xPos, int yPos, String name) {
		nodeList.add(new Destination(xPos, yPos, name, destIndex, nodeIndex));
		destinations.add(new Destination(xPos, yPos, name, destIndex, nodeIndex));
		destIndex++;
		nodeIndex++;
	}
	
	/**
	 * Makes a new destination at (xPos, yPos) using the default name.
	 * @param xPos the x coordinate of the new destination
	 * @param yPos the y coordinate of the new destination
	 */
	public void loadDestination(int xPos, int yPos) {
		nodeList.add(new Destination(xPos, yPos, destIndex, nodeIndex));
		destinations.add(new Destination(xPos, yPos, destIndex, nodeIndex));
		destIndex++;
		nodeIndex++;
	}
	
	/**
	 * Clears the frame and then draws all shapes, including the "Edit Mode" text if applicable.
	 */
	public void draw() {
		this.getFrame().clearGraphics();
		if (hasMap) {
			System.out.println("before adding map");
			this.getFrame().addObject(map);
			System.out.println("after adding map");
		}
		for (int c = 0; c < waypoints.size(); c++){
			this.getFrame().addObject(waypoints.get(c).getShape());
		}
		for (int c = 0; c < destinations.size(); c++){
			this.getFrame().addObject(destinations.get(c).getShape());
		}
		for (int c = 0; c < edges.size(); c++){
			this.getFrame().addObject(edges.get(c).getShape());
		}
		if (canEdit) {
			this.getFrame().addObject(new SGText(EDIT_TOGGLE_NAME, DISPLAY_X, DISPLAY_Y));
		}
	}
	
	/**
	 * This method finds the first created Node that is within 3 of the input coordinates.
	 * @param xPos the x position of the point
	 * @param yPos the y position of the point
	 * @return true if the coordinates are within 3 of a node, false otherwise
	 */
	public boolean isNearNode(int xPos, int yPos) {
		for (Node c : nodeList) {
			if (c.getX()+3 >= xPos && c.getX()-3 <= xPos &&
				c.getY()+3 >= yPos && c.getY()-3 <= yPos) {
				if (!hasStartNode) {
					startNode = c;
					hasStartNode = true;
				}
				else if (!hasEndNode) {
					endNode = c;
					hasEndNode = true;
				}
				return true;
			}
		}
		return false;
	}
	
	public boolean loadFile() {
		BufferedReader r = null;
		try {
			r = new BufferedReader(new FileReader(READ_FILE_NAME));
		} catch (FileNotFoundException e) {
			return epicFail();
		}

		try {
			int c = 0;
			int nodes = -1;
			while (true) {
				String line = r.readLine();
				if (line == null) {
					break;
				}
				String[] words = line.split("\t");
				int size = words.length;
				if (size == 1) {
					System.out.println("map instantiated as: "+words[0]);
					imageName = words[0];
					hasMap = true;
				}
				else if (words[0].equals(WP_PART)) {
				// Waypoint xPos yPos
					if (size == 3) {
						System.out.println("new waypoint");
						loadWaypoint(Integer.parseInt(words[1]), Integer.parseInt(words[2]));
						nodes++;
					}
					else {
						System.err.println("Bad waypoint in file at line "+c);
						return epicFail();
					}
				}
				else if (words[0].equals(DEST_PART)) {
					if (size == 3) {
					// Destination xPos yPos
						System.out.println("new destination");
						loadDestination(Integer.parseInt(words[1]), Integer.parseInt(words[2]));
						nodes++;
					}
					else if (size == 4) {
					// Destination xPos yPos name
						System.out.println("new destination");
						loadDestination(Integer.parseInt(words[1]), Integer.parseInt(words[2]), words[3]);
						nodes++;
					}
					else {
						System.err.println("Bad destination in file at line "+c);
						return epicFail();
					}
					
				}
				else if (words[0].equals(EDGE_PART)) {
				// Edge 1 3 237
					if (size == 4) {
						System.out.println("new edge");
						int node1 = Integer.parseInt(words[1]);
						int node2 = Integer.parseInt(words[2]);
						if (node1 == node2) {
							System.err.println("Edge tried to connect to itself at line "+c);
							return epicFail();
						}
						if (nodes < node1 || nodes < node2) {
							System.err.println("No such node in file at line "+c);
							return epicFail();
						}
						loadEdge(nodeList.get(node1), nodeList.get(node2), words[3]);
					}
					else {
						System.err.println("Bad edge in file at line "+c);
						return epicFail();
					}
				}
				
			c++;
			}
		} catch (IOException e) {
			return epicFail();
		}
		
		try {
			r.close();
		} catch (IOException e) {
			return epicFail();
		}
		return true;
	}
	
	public boolean saveFile() {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(WRITE_FILE_NAME));
			if (hasMap) {
				writer.write(imageName);
				writer.write(NEXT);
			}
			for (int c = 0; c < waypoints.size(); c++){
				writer.write(""+WP_PART+ TAB +waypoints.get(c).getX()+ TAB +waypoints.get(c).getY());
				writer.write(NEXT);
			}
			for (int c = 0; c < destinations.size(); c++){
				if (destinations.get(c).hasName()) {
					writer.write(""+DEST_PART+ TAB +destinations.get(c).getX()+ TAB +destinations.get(c).getY()+ TAB +destinations.get(c).getName());
					writer.write(NEXT);
				}
				else {
					writer.write(""+DEST_PART+ TAB +destinations.get(c).getX()+ TAB +destinations.get(c).getY());
					writer.write(NEXT);
				}
			}
			for (int c = 0; c < edges.size(); c++){
				writer.write(""+EDGE_PART+ TAB +edges.get(c).getStartNode().getNodeIndex()+ TAB +edges.get(c).getEndNode().getNodeIndex()+ TAB +edges.get(c).getCost());
				writer.write(NEXT);
			}
			writer.close();
		} catch (IOException e) {
			return false;
		}
		return true;
	}
	
	/**
	 * Dumps all data and returns false
	 * @return false
	 */
	public boolean epicFail() {
		System.err.println("Epic Fail");
		dumpData();
		return false;
	}
	
	/**
	 * Emptys all data structures
	 */
	private void dumpData() {
		while(waypoints.size() > 0)waypoints.remove(waypoints.size()-1);
		while(destinations.size() > 0)destinations.remove(destinations.size()-1);
		while(nodeList.size() > 0)nodeList.remove(nodeList.size()-1);
		while(edges.size() > 0)edges.remove(edges.size()-1);
		draw();
	}
	
	private ArrayList<Edge> findPath(Node origStart, Node origTarget) {
		int startIndex = origStart.getNodeIndex();
		int endIndex = origTarget.getNodeIndex();
		ArrayList<Node> nodesList = new ArrayList<Node>(nodeList);
		Node[] nodes = new Node[nodesList.size()];
		nodes = nodesList.toArray(nodes);
		for (int c = 0; c < nodeList.size(); c++) {
			nodes[c].setCost(Integer.MAX_VALUE);
			nodes[c].setPrevious(null);
		}
		Node start = nodes[startIndex];
		Node target = nodes[endIndex];
		start.setCost(0);
		while (!isNull(nodes)) {
			int smallCost = getSmallCostIndex(nodes);
			if (nodes[smallCost].getCost() == Integer.MAX_VALUE) {
				break;
			}
			Node thisNode = removeNode(smallCost, nodes);
			if (thisNode.equals(target)) {
				break;
			}
			for (Node otherNode : thisNode.getNodes()) {
				if (isIn(nodes, otherNode)) {
					double alt = thisNode.getCost() + getEdge(thisNode, otherNode).getCost();
					if (alt < otherNode.getCost()) {
						otherNode.setCost(alt);
						otherNode.setPrevious(thisNode);
					}
				}
			}
		}
		ArrayList<Edge> edges = new ArrayList<Edge>();
		Node thisNode = target;
		while (!thisNode.getPrevious().equals(null)) {
			edges.add(getEdge(thisNode, thisNode.getPrevious()));
			thisNode = thisNode.getPrevious();
		}
		return edges;
	}
	
	private Edge getEdge(Node thisNode, Node otherNode) {
		//TODO: this method
		return null;
	}

	private boolean isIn(Node[] nodes, Node otherNode) {
		for (Node node: nodes) {
			if (otherNode.equals(node)) {
				return true;
			}
		}
		return false;
	}

	private int getSmallCostIndex(Node[] nodes) {
		double smallestCost = Integer.MAX_VALUE;
		int smallIndex = -1;
		for (Node node : nodes) {
			if (!node.equals(null)) {
				if (node.getCost() < smallestCost) {
					smallestCost = node.getCost();
					smallIndex = node.getNodeIndex();
				}
			}
		}
		return smallIndex;
	}

	private boolean isNull(Node[] nodes) {
		for (Node node: nodes) {
			if (!node.equals(null)) {
				return false;
			}
		}
		return true;
	}
	
	private Node removeNode(int index, Node[] nodes) {
		Node ret = nodes[index];
		nodes[index] = null;
		return ret;
	}
}
